program Lexer;
{$H+,R+,Q+}
{$APPTYPE CONSOLE}

uses
  SysUtils,Math;

const
  KeywordsCount = 46;
  AutomateCount = KeywordsCount + 6;
  ASpaceNumber = KeywordsCount;
  ACommentNumber = KeywordsCount + 1;
  AIntegerNumber = KeywordsCount + 2;
  AStringNumber = KeywordsCount + 3;
  AIdentifierNumber = KeywordsCount + 4;
	ACharNumber = KeywordsCount + 5;
	
  IllegalJump = -1;  
  Flag =#0;
	LowAlphabet = #1;
  HighAlphabet = #255;
  UnknownValue = MaxInt;


type
  TNumber = integer;

  type TState = (stWorking, stOK, stFailed);

  TJumpes = array of array [char] of integer;
	PAutomate =^TAutomate;
  TAutomate = record
    Jumpes: TJumpes;
    Symbol: string;
  end;

  TWorkingAutomate=record
  	state:TState;
    position:integer;
    automate:PAutomate;
  end;

  TIdentifier = record
    Name: string;
    Value: TNumber;
  end;
  Token=record
  	name:string;
    value:integer;
  end;


  Tokens=record
  	ArrToken:array of Token;
    size:integer;
  end;

var
  ASpace, AInteger, AString, AIdentifier, AComment,AChar: TAutomate;
  AKeyword: array [0..KeywordsCount - 1] of TAutomate;
 	proga:string;
  Position:integer;

  //CommandNumber: integer;

  IdentifierCount: integer;
  Identifiers: array of TIdentifier;
  StringCount: integer;
  Strings: array of string;



function LoadFromFile(filename:string):String;
var
	inf:TextFile;
  s:String;
begin
	assign(inf,filename);
  reset(inf);
  Result:='';
  while not eof(inf) do begin
  	readln(inf,s);
  	if (pos('#include',s)=1) then begin
  		delete(s,1,9);
  		result:= result + LoadFromFile(s) + #10;
  	end
  	else begin
	    Result:=Result+s+#10;
	 	end;
  end;
  CloseFile(inf);
  Position:=0;
end;

function nextChar:char;
begin
	inc(Position);
//  curChar:=proga[Position];
  Result:=proga[Position];
end;

function IntToChars(x:integer):string;
begin
  Result := Chr(Byte(x shr 24)) + Chr(Byte(x shr 16)) +
   Chr(Byte(x shr 8)) + Chr(Byte(x));
end;
function CompareStrings(s1,s2:string):boolean;
begin
	result:=(AnsiUpperCase(s1)=AnsiUpperCase(s2));
end;

function CompareChars(c1,c2:Char):boolean;
begin
	result:=(AnsiUpperCase(c1)=AnsiUpperCase(c2));
end;
procedure BuildKeywordJumpes(Keyword: string; var Jumpes: TJumpes);
var
  l, i: integer;
  j: char;
begin
  l := Length(Keyword);
  SetLength(Jumpes, l + 1);
  for i := 0 to l - 1 do
  begin
    Jumpes[i, Flag] := 0;
    for j := LowAlphabet to HighAlphabet do
      if CompareChars(Keyword[i + 1], j) then
        Jumpes[i, j] := i + 1
      else
        Jumpes[i, j] := IllegalJump;
  end;
  Jumpes[l, Flag] := 1;
  for j := LowAlphabet to HighAlphabet do
    Jumpes[l, j] := IllegalJump;
end;

procedure GenerateKeywordAutomate(Symbol,Keyword: string; var AKeyword: TAutomate);
begin
  AKeyword.Symbol := Symbol;
  BuildKeywordJumpes(Keyword, AKeyword.Jumpes);
end;

procedure InitializeAutomates;
var
  i: char;
begin
  with ASpace do
  begin
    Symbol := 'ERROR SPACE IS NOT PROCESSED';
    SetLength(Jumpes, 2);
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        #0..#32: Jumpes[0, i] := 1;
        else     Jumpes[0, i] := IllegalJump;
      end;
    Jumpes[1, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      Jumpes[1, i] := IllegalJump;
  end;

  with AComment do
  begin
    Symbol := 'ERROR COMMENT IS NOT PROCESSED';
    SetLength(Jumpes, 7);
    
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '(': Jumpes[0, i] := 1;
        '/': Jumpes[0, i] := 5;
        else Jumpes[0, i] := IllegalJump;
      end;
    
    Jumpes[1, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '*': Jumpes[1, i] := 2;
        else Jumpes[1, i] := IllegalJump;
      end;
    
    Jumpes[2, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '*': Jumpes[2, i] := 3;
        else Jumpes[2, i] := 2;
      end;
    
    Jumpes[3, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '/',')': Jumpes[3, i] := 4;
        '*' : Jumpes[3, i] := 3;
        else Jumpes[3, i] := 2;
      end;
    
    Jumpes[4, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      Jumpes[4, i] := IllegalJump;
    
    Jumpes[5, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '*': Jumpes[5, i] := 2;
        '/': Jumpes[5, i] := 6;
        else Jumpes[5, i] := IllegalJump;
      end;
    
    Jumpes[6, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        #10,#13: Jumpes[6, i] := 4;
        else Jumpes[6, i] := 6;
      end;
  end;


  with AInteger do
  begin
    Symbol := 'num';
    SetLength(Jumpes, 2);
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '0'..'9': Jumpes[0, i] := 1;
        else      Jumpes[0, i] := IllegalJump;
      end;
    Jumpes[1, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '0'..'9': Jumpes[1, i] := 1;
        else      Jumpes[1, i] := IllegalJump;
      end;
  end;

  with AString do
  begin
    Symbol := 'str';
    SetLength(Jumpes, 3);
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '"': Jumpes[0, i] := 1;
        else  Jumpes[0, i] := IllegalJump;
      end;
    Jumpes[1, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '"': Jumpes[1, i] := 2;
        else  Jumpes[1, i] := 1;
      end;
    Jumpes[2, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      Jumpes[2, i] := IllegalJump;
  end;

  with AIdentifier do
  begin
    Symbol := 'id';
    SetLength(Jumpes, 2);
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        'A'..'Z', 'a'..'z', ''..'', ''..'', '', '', '_':
              Jumpes[0, i] := 1;
        else  Jumpes[0, i] := IllegalJump;
      end;
    Jumpes[1, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      case i of
        'A'..'Z', 'a'..'z', ''..'', ''..'', '', '', '_', '0'..'9':
              Jumpes[1, i] := 1;
        else  Jumpes[1, i] := IllegalJump;
      end;
  end;
  
  with AChar do
  begin
    Symbol := 'num';
    SetLength(Jumpes, 4);
    Jumpes[0, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '''': Jumpes[0, i] := 1;
        else  Jumpes[0, i] := IllegalJump;
      end;

    Jumpes[1, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do      
      Jumpes[1, i] := 2;

    Jumpes[2, Flag] := 0;
    for i := LowAlphabet to HighAlphabet do
      case i of
        '''': Jumpes[2, i] := 3;
        else  Jumpes[2, i] := IllegalJump;
      end;
    Jumpes[3, Flag] := 1;
    for i := LowAlphabet to HighAlphabet do
      Jumpes[3, i] := IllegalJump;
  end;

  GenerateKeywordAutomate('~', '=', AKeyword[0]);
  GenerateKeywordAutomate('+', '+', AKeyword[1]);
  GenerateKeywordAutomate('-', '-', AKeyword[2]);
  GenerateKeywordAutomate('*', '*', AKeyword[3]);
  GenerateKeywordAutomate('/', '/', AKeyword[4]);
  GenerateKeywordAutomate('/', 'div', AKeyword[5]);
  GenerateKeywordAutomate('%', '%', AKeyword[6]);
  GenerateKeywordAutomate('%', 'mod', AKeyword[7]);
  GenerateKeywordAutomate('^', '^', AKeyword[8]);
  GenerateKeywordAutomate('(', '(', AKeyword[9]);
  GenerateKeywordAutomate(')', ')', AKeyword[10]);
  GenerateKeywordAutomate('=', '==', AKeyword[11]);
  GenerateKeywordAutomate('<', '<', AKeyword[12]);
  GenerateKeywordAutomate('>', '>', AKeyword[13]);
  GenerateKeywordAutomate('!=', '<>', AKeyword[14]);
  GenerateKeywordAutomate('!=', '!=', AKeyword[15]);
  GenerateKeywordAutomate('{', '{', AKeyword[16]);
  GenerateKeywordAutomate('}', '}', AKeyword[17]);
  GenerateKeywordAutomate('{', 'begin', AKeyword[18]);
  GenerateKeywordAutomate('}', 'end', AKeyword[19]);
  GenerateKeywordAutomate(',', ',', AKeyword[20]);
  GenerateKeywordAutomate(';', ';', AKeyword[21]);
  GenerateKeywordAutomate('if', 'if', AKeyword[22]);
  GenerateKeywordAutomate('while', 'while', AKeyword[23]);
  GenerateKeywordAutomate('read', 'read', AKeyword[24]);
  GenerateKeywordAutomate('write', 'write', AKeyword[25]);
  GenerateKeywordAutomate('for', 'for', AKeyword[26]);
  GenerateKeywordAutomate('++', '++', AKeyword[27]);
  GenerateKeywordAutomate('--', '--', AKeyword[28]);
  GenerateKeywordAutomate('+=', '+=', AKeyword[29]);
  GenerateKeywordAutomate('-=', '-=', AKeyword[30]);
  GenerateKeywordAutomate('*=', '*=', AKeyword[31]);
  GenerateKeywordAutomate('/=', '/=', AKeyword[32]);
  GenerateKeywordAutomate('>=', '>=', AKeyword[33]);
  GenerateKeywordAutomate('<=', '<=', AKeyword[34]);
  GenerateKeywordAutomate('void', 'void', AKeyword[35]);
  GenerateKeywordAutomate('return', 'return', AKeyword[36]);
  GenerateKeywordAutomate('halt', 'halt', AKeyword[37]);
  GenerateKeywordAutomate('local', '&', AKeyword[38]);
  GenerateKeywordAutomate('int', 'int', AKeyword[39]);
  GenerateKeywordAutomate('[', '[', AKeyword[40]);
  GenerateKeywordAutomate(']', ']', AKeyword[41]);
  GenerateKeywordAutomate('readc', 'readc', AKeyword[42]);
  GenerateKeywordAutomate('writec', 'writec', AKeyword[43]);
  GenerateKeywordAutomate('pointer', '@', AKeyword[44]);
  GenerateKeywordAutomate('~', '~', AKeyword[45]);

	//TODO
	//change const
	//  KeywordsCount = last+1;  
	//  at line 8
end;

function PreCompile(s: string; var t:Tokens): integer;
var
  WorkingAutomates: array [0..AutomateCount - 1] of TWorkingAutomate;
  i,n:integer;
  p,StartP:integer;
  ThisOK,LastOK:Integer;
  AllFailed,EndOfString:Boolean;
begin
	IdentifierCount:=0;
  StringCount:=0;
	InitializeAutomates;
	for i:=0 to KeywordsCount-1 do begin
  	WorkingAutomates[i].automate:=@AKeyWord[i];
  end;
  WorkingAutomates[ASpaceNumber].Automate := @ASpace;
  WorkingAutomates[ACommentNumber].Automate := @AComment;
  WorkingAutomates[AIntegerNumber].Automate := @AInteger;
  WorkingAutomates[AStringNumber].Automate := @AString;
  WorkingAutomates[AIdentifierNumber].Automate := @AIdentifier;
  WorkingAutomates[ACharNumber].Automate := @AChar;
  for i := 0 to AutomateCount - 1 do begin
    WorkingAutomates[i].state := stWorking;
    WorkingAutomates[i].position := 0;
  end;
  t.size:=0;
  SetLength(t.ArrToken,2);
  p:=1;
  ThisOK:=-1;
  LastOK:=-1;
  AllFailed := true;  //{$O-}
  //StartP:=pos('+',s);
  StartP:=1;
  EndOfString := false;

  while true do begin
	  if p <= Length(s) then begin
      AllFailed := true;
      LastOK := ThisOK;
      ThisOK := -1;
      for i := 0 to AutomateCount - 1 do begin
        if WorkingAutomates[i].State = stFailed then Continue;
        if WorkingAutomates[i].Automate^.Jumpes[WorkingAutomates[i].Position, s[p]] = -1 then begin
          WorkingAutomates[i].State := stFailed;
          Continue;
        end;
        WorkingAutomates[i].Position := WorkingAutomates[i].Automate^.Jumpes[WorkingAutomates[i].Position, s[p]];
        if WorkingAutomates[i].Automate^.Jumpes[WorkingAutomates[i].Position, Flag] = 1 then begin
          WorkingAutomates[i].State := stOK;
          if ThisOK = -1 then
            ThisOK := i;
        end
        else begin
          WorkingAutomates[i].State := stWorking;
        end;
        AllFailed := false;
      end;
    end
    else
      EndOfString := true;

   if AllFailed or EndOfString then begin
      if EndOfString then
      	LastOK := ThisOK;
      if (LastOK = -1) then begin
        writeln(output,'error near (',startP,'..',p,') ',Copy(s, Max(StartP-10,1), p - StartP+20));
        Result := 1;
        Exit;
      end;
      inc(t.size);
   		SetLength(t.ArrToken,t.size+1);
      t.ArrToken[t.size].name:= WorkingAutomates[LastOK].Automate^.Symbol;
      case LastOK of
      	ASpaceNumber,ACommentNumber:
        begin
        	dec(t.size);
        end;
        AIntegerNumber:
        begin
          //Inc(NumberCount);
          //SetLength(Numbers, NumberCount);
          //Numbers[NumberCount - 1] := StrToInt(Copy(s, StartP, p - StartP));
          //t := t + IntToChars(NumberCount - 1);
          t.ArrToken[t.size].value:=StrToInt(Copy(s, StartP, p - StartP));
        end;
        AIdentifierNumber:
        begin
          n := IdentifierCount;
          for i := 0 to IdentifierCount - 1 do
            if CompareStrings(Identifiers[i].Name,Copy(s, StartP, p - StartP)) then begin
              n := i;
              Break;
            end;
          if n = IdentifierCount then begin
            Inc(IdentifierCount);
            SetLength(Identifiers, IdentifierCount);
            Identifiers[n].Name := Copy(s, StartP, p - StartP);
          end;
          Identifiers[n].Value := UnknownValue;
          t.ArrToken[t.size].value:=n;
        end;
        AStringNumber:
        begin
          Inc(StringCount);
          SetLength(Strings, StringCount);
          Strings[StringCount - 1] := Copy(s, StartP + 1, p - StartP - 2);
          t.ArrToken[t.size].value:=StringCount-1
        end;
        ACharNumber:
        begin
					t.ArrToken[t.size].value:=Ord(s[StartP+1]);        	
        end;
      end;

      for i := 0 to AutomateCount - 1 do
        with WorkingAutomates[i] do
        begin
          State := stWorking;
          Position := 0;
        end;
      StartP := p;
      Dec(p);
    end;

    if EndOfString then Break;
    Inc(p);
  end;
  result:=0;
end;



var
	LeX_proga:Tokens;
  i:integer;
  outputfile:textfile;
begin
	if ParamCount<>1 then begin
		writeln('using: Lexer "FileName"');
		halt(1);
	end;
	proga:=LoadFromFile(ParamStr(1));
  if PreCompile(proga,LeX_proga)<>0 then begin
  	writeln('SYNTAX ERROR');
    halt(1);
  end;
  assign(outputFile,ParamStr(1)+'.lex');
  rewrite(outputFile);
  writeln(outputfile,StringCount,' ',IdentifierCount);
  for i:=0 to StringCount-1 do begin
  	writeln(outputfile,Strings[i]);
  end;
  for i:=1 to LeX_proga.size do begin
  	writeln(outputfile,LeX_proga.ArrToken[i].name,#$FD,LeX_proga.ArrToken[i].value);
  	if (LeX_proga.ArrToken[i].name='}') then begin
  		writeln(outputfile,';',#$FD,0);
  	end;
  end;
  close(outputfile);
//  Writeln('*********"SYNTAX OK"*********');
//  readln;
end.
