program Runner;
{$APPTYPE CONSOLE}
{$DEFINE OLYMP}
{-$UNDEF OLYMP}
{$DEFINE DEBUG}
{$UNDEF DEBUG}
{$Q+,R+,S+}
uses
	SysUtils,Math;

type 
PArray = ^TArray;
TArray = record
	len:integer;
	values:array of Int64;
end;

local = record
	returnLine :integer;
	localVariable : array of Int64;
	localArray : array of TArray;
end;


var
	code:array of string;
  countLine,Line,IdCount:integer;
  stack,Id:array of Int64;
  _Array:array of TArray;
  voidStack : array of local;
	head:integer=0;
	sizeVoidStack:integer;

procedure push(x:Int64);
begin
	inc(head);
  SetLength(stack,head+1);
  stack[head]:=x;
end;

function isIdentifier(s:string):boolean;
begin
	result:=(s[1]='$');
end;

function isLocal(s:string):boolean;
begin
	result:=(s[2]='&') or ((Length(s)>=3) and (s[3]= '&'));
end;

function isPointer(s:string):boolean;
begin
	result:=(s[2]='@') or ((Length(s)>=3) and (s[3]= '@'));
end;

procedure Fail(msg:string);
begin
	writeln('RUNTIME ERROR: '+msg);
  halt(1);
end;


function StrToId(sid:string):Int64;
begin
	if not isIdentifier(sid) then fail('StrToId fail at line '+IntToStr(line));
  delete(sid,1,1);
  while sid[1] in ['&','@'] do delete(sid,1,1);
  result:=StrToInt64(sid);
end;

function StrToNum(sn:string):Int64;
begin
  result:=StrToInt64(sn);
end;

function pop:Int64;
begin
	if head<=0 then fail('stack underflow at line '+IntToStr(line));
	result:=stack[head];
  dec(head);
end;

function IntPower(base:Int64;power:Int64):Int64;
var i:integer;
begin
	if (base=0) and(power=0) then Fail('0^0 is not define');
	result:=1;
	for i:= 1 to power do begin
		result:=result*base;
	end;                  
end;

var
	InputFile:TextFile;
  s,op:string;
  x,y:Int64;
  boolResult:boolean=false;
	index :integer;
	variable :integer;
	c:char;
  i:integer;
  arr:PArray;
begin
	if ParamCount<>1 then begin
  	writeln('usage: Runner "FileName"');
    halt(1);
  end;
  assign(InputFile,ParamStr(1));
  reset(InputFile);
  readln(InputFile,IdCount);
  SetLength(Id,IdCount);
  SetLength(_Array,IdCount);
  for i:=0 to IdCount-1 do 
  	_Array[i].len:=0;

  countLine:=0;
  while not eof(InputFile) do begin
  	inc(countLine);
    SetLength(code,countLine+1);
    readln(InputFile,code[countLine]);
  end;
  CloseFile(InputFile);
  line:=1;
  sizeVoidStack:=0;
  while code[line]<>'stop' do begin
  	s:=code[line];
  	{$IFDEF DEBUG}
	  	write(erroutput,s,' head=',head,' ');
	  	for i:=1 to head do write(erroutput,stack[i],' ');
	  	writeln(erroutput);
	  {$ENDIF}
    s:=s+' ';
    op:=copy(s,1,pos(' ',s)-1);
		delete(s,1,pos(' ',s));
		s:=Trim(s);
    if op='read' then begin
    	{$IFDEF OLYMP}
    		read(x);
    	{$ELSE}
    		readln(x);
    	{$ENDIF}
      push(x);
    end
    else if op='write' then begin
    	{$IFDEF OLYMP}
    		write(pop);
    	{$ELSE}
    		writeln(pop);
    	{$ENDIF}
    end
    else if op='pop' then begin
    	if s='' then pop else
    	if not isLocal(s) then 
    		Id[StrToId(s)]:=pop
    	else 
    		voidStack[sizeVoidStack-1].localVariable[StrToId(s)] := pop;

    end
    else if op='push' then begin
    	if isIdentifier(s) then begin
      	if (not isLocal(s)) then
      		push(Id[StrToId(s)])
      	else
      		push(voidStack[sizeVoidStack-1].localVariable[StrToId(s)]);
      end
      else begin
      	push(StrToNum(s));
      end;
    end
    else if op='mul' then begin
      push(pop*pop);
    end
    else if op='add' then begin
      push(pop+pop);
    end
    else if op='sub' then begin
    	x:=pop;
      push(pop-x);
    end
    else if op='div' then begin
    	x:=pop;
      push(pop div x);
    end
    else if op='mod' then begin
    	x:=pop;
      push(pop mod x);
    end
    else if op='pow' then begin
    	x := pop;
    	y := pop;
    	push(IntPower(y,x));
    end
    else if op='more' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x>y);
    end
    else if op='less' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x<y);
    end
    else if op='equal' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x=y);
    end
    else if op='noteq' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x<>y);
    end
    else if op='lesseq' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x<=y);
    end
    else if op='moreeq' then begin
    	y:=pop;
      x:=pop;
      boolResult:=(x>=y);
    end
    else if op='jfalse' then begin
    	if not boolResult then begin
      	line:=StrToInt(s);
        continue;
      end;
    end
    else if op='jump' then begin
     	line:=StrToInt(s);
      continue;
    end
    else if op='sign' then begin
    	push(-pop);
    end
    else if op='print' then begin
      delete(s,1,1);
      delete(s,length(s),1);
    	if s='/n' then writeln
      else write(s);
    end
    else if op='call' then begin
    	inc(SizevoidStack);
			SetLength(voidStack,sizeVoidStack);
			voidStack[sizeVoidStack-1].returnLine:=line+1;
			SetLength(voidStack[sizeVoidStack-1].localVariable,IdCount);
			SetLength(voidStack[sizeVoidStack-1].localArray,IdCount);
		  for i:=0 to IdCount-1 do 
  			voidStack[sizeVoidStack-1].localArray[i].len:=0;

			line := StrToInt(s);
			if (line<0) then Fail('fail in start void/int, jump to -1');
//			writeln(line);
			continue;
    end
    else if op='return' then begin
  //  	writeln('!',sizeVoidStack);
			line:=voidStack[sizeVoidStack-1].returnLine;
			dec(sizeVoidStack);
			SetLength(voidStack,sizeVoidStack);
			continue;
    end
    else if op='nop' then begin
    	//do nothing
    end
    else if op='pushpoppop' then begin
			x:=pop;
			push(x);
			push(x);			
    end
    else if op='arrpop' then begin
      x := pop;
   		index := Integer(pop);
   		variable := Integer(StrToId(s));
   		if isPointer(s) then begin
      	if not isLocal(s) then
	   			arr := PArray(Pointer(Integer(ID[variable])))
      	else
	   			arr := PArray(Pointer(Integer(voidStack[sizeVoidStack-1].localVariable[variable])));
				if arr^.len<=index then begin
					arr^.len := index + 1;
					SetLength(arr^.values,arr^.len);
	     	end;
	     	arr^.values[index]:=x;
   		end
   		else begin
	    	if not isLocal(s) then begin
  	   		if _Array[variable].len<=index then begin
						_Array[variable].len := index + 1;
						SetLength(_Array[variable].values,_Array[variable].len);
	     		end;
	     		_Array[variable].values[index]:=x;
	    	end
	    	else begin
	     		if voidStack[sizeVoidStack-1].localArray[variable].len<=index then begin
						voidStack[sizeVoidStack-1].localArray[variable].len := index + 1;
						SetLength(voidStack[sizeVoidStack-1].localArray[variable].values,voidStack[sizeVoidStack-1].localArray[variable].len);
	     		end;
	     		voidStack[sizeVoidStack-1].localArray[variable].values[index] := x;
  	  	end;
  	  end;

    end
    else if op='arrpush' then begin
   		index := Integer(pop);
   		variable := Integer(StrToId(s));
   		if isPointer(s) then begin
      	if not isLocal(s) then
	   			arr := PArray(Pointer(Integer(ID[variable])))
      	else
	   			arr := PArray(Pointer(Integer(voidStack[sizeVoidStack-1].localVariable[variable])));
//   			arr := PArray(Pointer(Integer(ID[variable])));
    		if arr^.len<=index then begin
					arr^.len := index + 1;
					SetLength(arr^.values,arr^.len);
     		end;
     		push(arr^.values[index])
   		end
   		else begin
	     	if not isLocal(s) then begin
	     		if _Array[variable].len<=index then begin
						_Array[variable].len := index + 1;
						SetLength(_Array[variable].values,_Array[variable].len);
	     		end;
	     		push(_Array[variable].values[index])
	     	end
	     	else begin
  	   		if voidStack[sizeVoidStack-1].localArray[variable].len<=index then begin
						voidStack[sizeVoidStack-1].localArray[variable].len := index + 1;
						SetLength(voidStack[sizeVoidStack-1].localArray[variable].values,voidStack[sizeVoidStack-1].localArray[variable].len);
	     		end;
	     		push(voidStack[sizeVoidStack-1].localArray[variable].values[index])
	     	end;
	  	end;
    end
    else if op='arrgetlen' then begin
   		variable := Integer(StrToId(s));
   		if isPointer(s) then begin
      	if not isLocal(s) then
	   			arr := PArray(Pointer(Integer(ID[variable])))
      	else
	   			arr := PArray(Pointer(Integer(voidStack[sizeVoidStack-1].localVariable[variable])));
//   			arr := PArray(Pointer(Integer(ID[variable])));
     		push(arr^.len)   			
   		end
   		else begin      
	     	if not isLocal(s) then begin
	     		push(_Array[variable].len);	
	     	end
	     	else begin
	     		push(voidStack[sizeVoidStack-1].localArray[variable].len);
  	   	end;
			end;	
    end
    else if op='writec' then begin
    	write(chr(Byte(pop and $FF)));
    end
    else if op='readc' then begin
			read(c);
			push(ord(c));
    end
    else if op='arrgetpointer' then begin
   		variable := Integer(StrToId(s));
     	if not isLocal(s) then begin
     		push(Int64(@_Array[variable]));
     	end
     	else begin
     		push(Int64(@voidStack[sizeVoidStack-1].localArray[variable]));
     	end;
    end
    else if op='' then begin

    end
    else if op='' then begin

    end
    else if op='' then begin

    end
    else if op='' then begin

    end
    else if op='' then begin

    end
    else Fail('Unknown operation "'+op+'"');
    inc(line);
  end;
	writeln;
end.